#include "linepoint.vert.metal.h"
#include "tri_color.vert.metal.h"
#include "tri_texture.vert.metal.h"
#include "color.frag.metal.h"
#include "texture_rgba.frag.metal.h"
#include "texture_rgb.frag.metal.h"
