#include "linepoint.vert.sm50.dxbc.h"
#include "tri_color.vert.sm50.dxbc.h"
#include "tri_texture.vert.sm50.dxbc.h"
#include "color.frag.sm50.dxbc.h"
#include "texture_rgba.frag.sm50.dxbc.h"
#include "texture_rgb.frag.sm50.dxbc.h"
