/*
  Simple DirectMedia Layer
  Copyright (C) 1997-2025 Sam Lantinga <slouken@libsdl.org>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/


// DO NOT EDIT THIS FILE BY HAND. It is autogenerated by gendynapi.py.

#if !SDL_DYNAMIC_API
#error You should not be here.
#endif

// New API symbols are added at the end
#define SDL_AcquireCameraFrame SDL_AcquireCameraFrame_REAL
#define SDL_AcquireGPUCommandBuffer SDL_AcquireGPUCommandBuffer_REAL
#define SDL_AcquireGPUSwapchainTexture SDL_AcquireGPUSwapchainTexture_REAL
#define SDL_AddAtomicInt SDL_AddAtomicInt_REAL
#define SDL_AddEventWatch SDL_AddEventWatch_REAL
#define SDL_AddGamepadMapping SDL_AddGamepadMapping_REAL
#define SDL_AddGamepadMappingsFromFile SDL_AddGamepadMappingsFromFile_REAL
#define SDL_AddGamepadMappingsFromIO SDL_AddGamepadMappingsFromIO_REAL
#define SDL_AddHintCallback SDL_AddHintCallback_REAL
#define SDL_AddSurfaceAlternateImage SDL_AddSurfaceAlternateImage_REAL
#define SDL_AddTimer SDL_AddTimer_REAL
#define SDL_AddTimerNS SDL_AddTimerNS_REAL
#define SDL_AddVulkanRenderSemaphores SDL_AddVulkanRenderSemaphores_REAL
#define SDL_AttachVirtualJoystick SDL_AttachVirtualJoystick_REAL
#define SDL_AudioDevicePaused SDL_AudioDevicePaused_REAL
#define SDL_BeginGPUComputePass SDL_BeginGPUComputePass_REAL
#define SDL_BeginGPUCopyPass SDL_BeginGPUCopyPass_REAL
#define SDL_BeginGPURenderPass SDL_BeginGPURenderPass_REAL
#define SDL_BindAudioStream SDL_BindAudioStream_REAL
#define SDL_BindAudioStreams SDL_BindAudioStreams_REAL
#define SDL_BindGPUComputePipeline SDL_BindGPUComputePipeline_REAL
#define SDL_BindGPUComputeSamplers SDL_BindGPUComputeSamplers_REAL
#define SDL_BindGPUComputeStorageBuffers SDL_BindGPUComputeStorageBuffers_REAL
#define SDL_BindGPUComputeStorageTextures SDL_BindGPUComputeStorageTextures_REAL
#define SDL_BindGPUFragmentSamplers SDL_BindGPUFragmentSamplers_REAL
#define SDL_BindGPUFragmentStorageBuffers SDL_BindGPUFragmentStorageBuffers_REAL
#define SDL_BindGPUFragmentStorageTextures SDL_BindGPUFragmentStorageTextures_REAL
#define SDL_BindGPUGraphicsPipeline SDL_BindGPUGraphicsPipeline_REAL
#define SDL_BindGPUIndexBuffer SDL_BindGPUIndexBuffer_REAL
#define SDL_BindGPUVertexBuffers SDL_BindGPUVertexBuffers_REAL
#define SDL_BindGPUVertexSamplers SDL_BindGPUVertexSamplers_REAL
#define SDL_BindGPUVertexStorageBuffers SDL_BindGPUVertexStorageBuffers_REAL
#define SDL_BindGPUVertexStorageTextures SDL_BindGPUVertexStorageTextures_REAL
#define SDL_BlitGPUTexture SDL_BlitGPUTexture_REAL
#define SDL_BlitSurface SDL_BlitSurface_REAL
#define SDL_BlitSurface9Grid SDL_BlitSurface9Grid_REAL
#define SDL_BlitSurfaceScaled SDL_BlitSurfaceScaled_REAL
#define SDL_BlitSurfaceTiled SDL_BlitSurfaceTiled_REAL
#define SDL_BlitSurfaceTiledWithScale SDL_BlitSurfaceTiledWithScale_REAL
#define SDL_BlitSurfaceUnchecked SDL_BlitSurfaceUnchecked_REAL
#define SDL_BlitSurfaceUncheckedScaled SDL_BlitSurfaceUncheckedScaled_REAL
#define SDL_BroadcastCondition SDL_BroadcastCondition_REAL
#define SDL_CaptureMouse SDL_CaptureMouse_REAL
#define SDL_ClaimWindowForGPUDevice SDL_ClaimWindowForGPUDevice_REAL
#define SDL_CleanupTLS SDL_CleanupTLS_REAL
#define SDL_ClearAudioStream SDL_ClearAudioStream_REAL
#define SDL_ClearClipboardData SDL_ClearClipboardData_REAL
#define SDL_ClearComposition SDL_ClearComposition_REAL
#define SDL_ClearError SDL_ClearError_REAL
#define SDL_ClearProperty SDL_ClearProperty_REAL
#define SDL_ClearSurface SDL_ClearSurface_REAL
#define SDL_CloseAudioDevice SDL_CloseAudioDevice_REAL
#define SDL_CloseCamera SDL_CloseCamera_REAL
#define SDL_CloseGamepad SDL_CloseGamepad_REAL
#define SDL_CloseHaptic SDL_CloseHaptic_REAL
#define SDL_CloseIO SDL_CloseIO_REAL
#define SDL_CloseJoystick SDL_CloseJoystick_REAL
#define SDL_CloseSensor SDL_CloseSensor_REAL
#define SDL_CloseStorage SDL_CloseStorage_REAL
#define SDL_CompareAndSwapAtomicInt SDL_CompareAndSwapAtomicInt_REAL
#define SDL_CompareAndSwapAtomicPointer SDL_CompareAndSwapAtomicPointer_REAL
#define SDL_CompareAndSwapAtomicU32 SDL_CompareAndSwapAtomicU32_REAL
#define SDL_ComposeCustomBlendMode SDL_ComposeCustomBlendMode_REAL
#define SDL_ConvertAudioSamples SDL_ConvertAudioSamples_REAL
#define SDL_ConvertEventToRenderCoordinates SDL_ConvertEventToRenderCoordinates_REAL
#define SDL_ConvertPixels SDL_ConvertPixels_REAL
#define SDL_ConvertPixelsAndColorspace SDL_ConvertPixelsAndColorspace_REAL
#define SDL_ConvertSurface SDL_ConvertSurface_REAL
#define SDL_ConvertSurfaceAndColorspace SDL_ConvertSurfaceAndColorspace_REAL
#define SDL_CopyFile SDL_CopyFile_REAL
#define SDL_CopyGPUBufferToBuffer SDL_CopyGPUBufferToBuffer_REAL
#define SDL_CopyGPUTextureToTexture SDL_CopyGPUTextureToTexture_REAL
#define SDL_CopyProperties SDL_CopyProperties_REAL
#define SDL_CopyStorageFile SDL_CopyStorageFile_REAL
#define SDL_CreateAudioStream SDL_CreateAudioStream_REAL
#define SDL_CreateColorCursor SDL_CreateColorCursor_REAL
#define SDL_CreateCondition SDL_CreateCondition_REAL
#define SDL_CreateCursor SDL_CreateCursor_REAL
#define SDL_CreateDirectory SDL_CreateDirectory_REAL
#define SDL_CreateEnvironment SDL_CreateEnvironment_REAL
#define SDL_CreateGPUBuffer SDL_CreateGPUBuffer_REAL
#define SDL_CreateGPUComputePipeline SDL_CreateGPUComputePipeline_REAL
#define SDL_CreateGPUDevice SDL_CreateGPUDevice_REAL
#define SDL_CreateGPUDeviceWithProperties SDL_CreateGPUDeviceWithProperties_REAL
#define SDL_CreateGPUGraphicsPipeline SDL_CreateGPUGraphicsPipeline_REAL
#define SDL_CreateGPUSampler SDL_CreateGPUSampler_REAL
#define SDL_CreateGPUShader SDL_CreateGPUShader_REAL
#define SDL_CreateGPUTexture SDL_CreateGPUTexture_REAL
#define SDL_CreateGPUTransferBuffer SDL_CreateGPUTransferBuffer_REAL
#define SDL_CreateHapticEffect SDL_CreateHapticEffect_REAL
#define SDL_CreateMutex SDL_CreateMutex_REAL
#define SDL_CreatePalette SDL_CreatePalette_REAL
#define SDL_CreatePopupWindow SDL_CreatePopupWindow_REAL
#define SDL_CreateProcess SDL_CreateProcess_REAL
#define SDL_CreateProcessWithProperties SDL_CreateProcessWithProperties_REAL
#define SDL_CreateProperties SDL_CreateProperties_REAL
#define SDL_CreateRWLock SDL_CreateRWLock_REAL
#define SDL_CreateRenderer SDL_CreateRenderer_REAL
#define SDL_CreateRendererWithProperties SDL_CreateRendererWithProperties_REAL
#define SDL_CreateSemaphore SDL_CreateSemaphore_REAL
#define SDL_CreateSoftwareRenderer SDL_CreateSoftwareRenderer_REAL
#define SDL_CreateStorageDirectory SDL_CreateStorageDirectory_REAL
#define SDL_CreateSurface SDL_CreateSurface_REAL
#define SDL_CreateSurfaceFrom SDL_CreateSurfaceFrom_REAL
#define SDL_CreateSurfacePalette SDL_CreateSurfacePalette_REAL
#define SDL_CreateSystemCursor SDL_CreateSystemCursor_REAL
#define SDL_CreateTexture SDL_CreateTexture_REAL
#define SDL_CreateTextureFromSurface SDL_CreateTextureFromSurface_REAL
#define SDL_CreateTextureWithProperties SDL_CreateTextureWithProperties_REAL
#define SDL_CreateThreadRuntime    SDL_CreateThreadRuntime_REAL
#define SDL_CreateThreadWithPropertiesRuntime   SDL_CreateThreadWithPropertiesRuntime_REAL
#define SDL_CreateWindow SDL_CreateWindow_REAL
#define SDL_CreateWindowAndRenderer SDL_CreateWindowAndRenderer_REAL
#define SDL_CreateWindowWithProperties SDL_CreateWindowWithProperties_REAL
#define SDL_CursorVisible SDL_CursorVisible_REAL
#define SDL_DateTimeToTime SDL_DateTimeToTime_REAL
#define SDL_Delay SDL_Delay_REAL
#define SDL_DelayNS SDL_DelayNS_REAL
#define SDL_DestroyAudioStream SDL_DestroyAudioStream_REAL
#define SDL_DestroyCondition SDL_DestroyCondition_REAL
#define SDL_DestroyCursor SDL_DestroyCursor_REAL
#define SDL_DestroyEnvironment SDL_DestroyEnvironment_REAL
#define SDL_DestroyGPUDevice SDL_DestroyGPUDevice_REAL
#define SDL_DestroyHapticEffect SDL_DestroyHapticEffect_REAL
#define SDL_DestroyMutex SDL_DestroyMutex_REAL
#define SDL_DestroyPalette SDL_DestroyPalette_REAL
#define SDL_DestroyProcess SDL_DestroyProcess_REAL
#define SDL_DestroyProperties SDL_DestroyProperties_REAL
#define SDL_DestroyRWLock SDL_DestroyRWLock_REAL
#define SDL_DestroyRenderer SDL_DestroyRenderer_REAL
#define SDL_DestroySemaphore SDL_DestroySemaphore_REAL
#define SDL_DestroySurface SDL_DestroySurface_REAL
#define SDL_DestroyTexture SDL_DestroyTexture_REAL
#define SDL_DestroyWindow SDL_DestroyWindow_REAL
#define SDL_DestroyWindowSurface SDL_DestroyWindowSurface_REAL
#define SDL_DetachThread SDL_DetachThread_REAL
#define SDL_DetachVirtualJoystick SDL_DetachVirtualJoystick_REAL
#define SDL_DisableScreenSaver SDL_DisableScreenSaver_REAL
#define SDL_DispatchGPUCompute SDL_DispatchGPUCompute_REAL
#define SDL_DispatchGPUComputeIndirect SDL_DispatchGPUComputeIndirect_REAL
#define SDL_DownloadFromGPUBuffer SDL_DownloadFromGPUBuffer_REAL
#define SDL_DownloadFromGPUTexture SDL_DownloadFromGPUTexture_REAL
#define SDL_DrawGPUIndexedPrimitives SDL_DrawGPUIndexedPrimitives_REAL
#define SDL_DrawGPUIndexedPrimitivesIndirect SDL_DrawGPUIndexedPrimitivesIndirect_REAL
#define SDL_DrawGPUPrimitives SDL_DrawGPUPrimitives_REAL
#define SDL_DrawGPUPrimitivesIndirect SDL_DrawGPUPrimitivesIndirect_REAL
#define SDL_DuplicateSurface SDL_DuplicateSurface_REAL
#define SDL_EGL_GetCurrentConfig SDL_EGL_GetCurrentConfig_REAL
#define SDL_EGL_GetCurrentDisplay SDL_EGL_GetCurrentDisplay_REAL
#define SDL_EGL_GetProcAddress SDL_EGL_GetProcAddress_REAL
#define SDL_EGL_GetWindowSurface SDL_EGL_GetWindowSurface_REAL
#define SDL_EGL_SetAttributeCallbacks SDL_EGL_SetAttributeCallbacks_REAL
#define SDL_EnableScreenSaver SDL_EnableScreenSaver_REAL
#define SDL_EndGPUComputePass SDL_EndGPUComputePass_REAL
#define SDL_EndGPUCopyPass SDL_EndGPUCopyPass_REAL
#define SDL_EndGPURenderPass SDL_EndGPURenderPass_REAL
#define SDL_EnterAppMainCallbacks SDL_EnterAppMainCallbacks_REAL
#define SDL_EnumerateDirectory SDL_EnumerateDirectory_REAL
#define SDL_EnumerateProperties SDL_EnumerateProperties_REAL
#define SDL_EnumerateStorageDirectory SDL_EnumerateStorageDirectory_REAL
#define SDL_EventEnabled SDL_EventEnabled_REAL
#define SDL_FillSurfaceRect SDL_FillSurfaceRect_REAL
#define SDL_FillSurfaceRects SDL_FillSurfaceRects_REAL
#define SDL_FilterEvents SDL_FilterEvents_REAL
#define SDL_FlashWindow SDL_FlashWindow_REAL
#define SDL_FlipSurface SDL_FlipSurface_REAL
#define SDL_FlushAudioStream SDL_FlushAudioStream_REAL
#define SDL_FlushEvent SDL_FlushEvent_REAL
#define SDL_FlushEvents SDL_FlushEvents_REAL
#define SDL_FlushIO SDL_FlushIO_REAL
#define SDL_FlushRenderer SDL_FlushRenderer_REAL
#define SDL_GDKResumeGPU SDL_GDKResumeGPU_REAL
#define SDL_GDKSuspendComplete SDL_GDKSuspendComplete_REAL
#define SDL_GDKSuspendGPU SDL_GDKSuspendGPU_REAL
#define SDL_GL_CreateContext SDL_GL_CreateContext_REAL
#define SDL_GL_DestroyContext SDL_GL_DestroyContext_REAL
#define SDL_GL_ExtensionSupported SDL_GL_ExtensionSupported_REAL
#define SDL_GL_GetAttribute SDL_GL_GetAttribute_REAL
#define SDL_GL_GetCurrentContext SDL_GL_GetCurrentContext_REAL
#define SDL_GL_GetCurrentWindow SDL_GL_GetCurrentWindow_REAL
#define SDL_GL_GetProcAddress SDL_GL_GetProcAddress_REAL
#define SDL_GL_GetSwapInterval SDL_GL_GetSwapInterval_REAL
#define SDL_GL_LoadLibrary SDL_GL_LoadLibrary_REAL
#define SDL_GL_MakeCurrent SDL_GL_MakeCurrent_REAL
#define SDL_GL_ResetAttributes SDL_GL_ResetAttributes_REAL
#define SDL_GL_SetAttribute SDL_GL_SetAttribute_REAL
#define SDL_GL_SetSwapInterval SDL_GL_SetSwapInterval_REAL
#define SDL_GL_SwapWindow SDL_GL_SwapWindow_REAL
#define SDL_GL_UnloadLibrary SDL_GL_UnloadLibrary_REAL
#define SDL_GPUSupportsProperties SDL_GPUSupportsProperties_REAL
#define SDL_GPUSupportsShaderFormats SDL_GPUSupportsShaderFormats_REAL
#define SDL_GPUTextureFormatTexelBlockSize SDL_GPUTextureFormatTexelBlockSize_REAL
#define SDL_GPUTextureSupportsFormat SDL_GPUTextureSupportsFormat_REAL
#define SDL_GPUTextureSupportsSampleCount SDL_GPUTextureSupportsSampleCount_REAL
#define SDL_GUIDToString SDL_GUIDToString_REAL
#define SDL_GamepadConnected SDL_GamepadConnected_REAL
#define SDL_GamepadEventsEnabled SDL_GamepadEventsEnabled_REAL
#define SDL_GamepadHasAxis SDL_GamepadHasAxis_REAL
#define SDL_GamepadHasButton SDL_GamepadHasButton_REAL
#define SDL_GamepadHasSensor SDL_GamepadHasSensor_REAL
#define SDL_GamepadSensorEnabled SDL_GamepadSensorEnabled_REAL
#define SDL_GenerateMipmapsForGPUTexture SDL_GenerateMipmapsForGPUTexture_REAL
#define SDL_GetAndroidActivity SDL_GetAndroidActivity_REAL
#define SDL_GetAndroidCachePath SDL_GetAndroidCachePath_REAL
#define SDL_GetAndroidExternalStoragePath SDL_GetAndroidExternalStoragePath_REAL
#define SDL_GetAndroidExternalStorageState SDL_GetAndroidExternalStorageState_REAL
#define SDL_GetAndroidInternalStoragePath SDL_GetAndroidInternalStoragePath_REAL
#define SDL_GetAndroidJNIEnv SDL_GetAndroidJNIEnv_REAL
#define SDL_GetAndroidSDKVersion SDL_GetAndroidSDKVersion_REAL
#define SDL_GetAppMetadataProperty SDL_GetAppMetadataProperty_REAL
#define SDL_GetAssertionHandler SDL_GetAssertionHandler_REAL
#define SDL_GetAssertionReport SDL_GetAssertionReport_REAL
#define SDL_GetAtomicInt SDL_GetAtomicInt_REAL
#define SDL_GetAtomicPointer SDL_GetAtomicPointer_REAL
#define SDL_GetAtomicU32 SDL_GetAtomicU32_REAL
#define SDL_GetAudioDeviceChannelMap SDL_GetAudioDeviceChannelMap_REAL
#define SDL_GetAudioDeviceFormat SDL_GetAudioDeviceFormat_REAL
#define SDL_GetAudioDeviceGain SDL_GetAudioDeviceGain_REAL
#define SDL_GetAudioDeviceName SDL_GetAudioDeviceName_REAL
#define SDL_GetAudioDriver SDL_GetAudioDriver_REAL
#define SDL_GetAudioFormatName SDL_GetAudioFormatName_REAL
#define SDL_GetAudioPlaybackDevices SDL_GetAudioPlaybackDevices_REAL
#define SDL_GetAudioRecordingDevices SDL_GetAudioRecordingDevices_REAL
#define SDL_GetAudioStreamAvailable SDL_GetAudioStreamAvailable_REAL
#define SDL_GetAudioStreamData SDL_GetAudioStreamData_REAL
#define SDL_GetAudioStreamDevice SDL_GetAudioStreamDevice_REAL
#define SDL_GetAudioStreamFormat SDL_GetAudioStreamFormat_REAL
#define SDL_GetAudioStreamFrequencyRatio SDL_GetAudioStreamFrequencyRatio_REAL
#define SDL_GetAudioStreamGain SDL_GetAudioStreamGain_REAL
#define SDL_GetAudioStreamInputChannelMap SDL_GetAudioStreamInputChannelMap_REAL
#define SDL_GetAudioStreamOutputChannelMap SDL_GetAudioStreamOutputChannelMap_REAL
#define SDL_GetAudioStreamProperties SDL_GetAudioStreamProperties_REAL
#define SDL_GetAudioStreamQueued SDL_GetAudioStreamQueued_REAL
#define SDL_GetBasePath SDL_GetBasePath_REAL
#define SDL_GetBooleanProperty SDL_GetBooleanProperty_REAL
#define SDL_GetCPUCacheLineSize SDL_GetCPUCacheLineSize_REAL
#define SDL_GetCameraDriver SDL_GetCameraDriver_REAL
#define SDL_GetCameraFormat SDL_GetCameraFormat_REAL
#define SDL_GetCameraID SDL_GetCameraID_REAL
#define SDL_GetCameraName SDL_GetCameraName_REAL
#define SDL_GetCameraPermissionState SDL_GetCameraPermissionState_REAL
#define SDL_GetCameraPosition SDL_GetCameraPosition_REAL
#define SDL_GetCameraProperties SDL_GetCameraProperties_REAL
#define SDL_GetCameraSupportedFormats SDL_GetCameraSupportedFormats_REAL
#define SDL_GetCameras SDL_GetCameras_REAL
#define SDL_GetClipboardData SDL_GetClipboardData_REAL
#define SDL_GetClipboardMimeTypes SDL_GetClipboardMimeTypes_REAL
#define SDL_GetClipboardText SDL_GetClipboardText_REAL
#define SDL_GetClosestFullscreenDisplayMode SDL_GetClosestFullscreenDisplayMode_REAL
#define SDL_GetCurrentAudioDriver SDL_GetCurrentAudioDriver_REAL
#define SDL_GetCurrentCameraDriver SDL_GetCurrentCameraDriver_REAL
#define SDL_GetCurrentDisplayMode SDL_GetCurrentDisplayMode_REAL
#define SDL_GetCurrentDisplayOrientation SDL_GetCurrentDisplayOrientation_REAL
#define SDL_GetCurrentRenderOutputSize SDL_GetCurrentRenderOutputSize_REAL
#define SDL_GetCurrentThreadID SDL_GetCurrentThreadID_REAL
#define SDL_GetCurrentTime SDL_GetCurrentTime_REAL
#define SDL_GetCurrentVideoDriver SDL_GetCurrentVideoDriver_REAL
#define SDL_GetCursor SDL_GetCursor_REAL
#define SDL_GetDXGIOutputInfo   SDL_GetDXGIOutputInfo_REAL
#define SDL_GetDateTimeLocalePreferences SDL_GetDateTimeLocalePreferences_REAL
#define SDL_GetDayOfWeek SDL_GetDayOfWeek_REAL
#define SDL_GetDayOfYear SDL_GetDayOfYear_REAL
#define SDL_GetDaysInMonth SDL_GetDaysInMonth_REAL
#define SDL_GetDefaultAssertionHandler SDL_GetDefaultAssertionHandler_REAL
#define SDL_GetDefaultCursor SDL_GetDefaultCursor_REAL
#define SDL_GetDesktopDisplayMode SDL_GetDesktopDisplayMode_REAL
#define SDL_GetDirect3D9AdapterIndex SDL_GetDirect3D9AdapterIndex_REAL
#define SDL_GetDisplayBounds SDL_GetDisplayBounds_REAL
#define SDL_GetDisplayContentScale SDL_GetDisplayContentScale_REAL
#define SDL_GetDisplayForPoint SDL_GetDisplayForPoint_REAL
#define SDL_GetDisplayForRect SDL_GetDisplayForRect_REAL
#define SDL_GetDisplayForWindow SDL_GetDisplayForWindow_REAL
#define SDL_GetDisplayName SDL_GetDisplayName_REAL
#define SDL_GetDisplayProperties SDL_GetDisplayProperties_REAL
#define SDL_GetDisplayUsableBounds SDL_GetDisplayUsableBounds_REAL
#define SDL_GetDisplays SDL_GetDisplays_REAL
#define SDL_GetEnvironment SDL_GetEnvironment_REAL
#define SDL_GetEnvironmentVariable SDL_GetEnvironmentVariable_REAL
#define SDL_GetEnvironmentVariables SDL_GetEnvironmentVariables_REAL
#define SDL_GetError SDL_GetError_REAL
#define SDL_GetEventFilter SDL_GetEventFilter_REAL
#define SDL_GetFloatProperty SDL_GetFloatProperty_REAL
#define SDL_GetFullscreenDisplayModes SDL_GetFullscreenDisplayModes_REAL
#define SDL_GetGDKDefaultUser SDL_GetGDKDefaultUser_REAL
#define SDL_GetGDKTaskQueue SDL_GetGDKTaskQueue_REAL
#define SDL_GetGPUDeviceDriver SDL_GetGPUDeviceDriver_REAL
#define SDL_GetGPUDriver SDL_GetGPUDriver_REAL
#define SDL_GetGPUShaderFormats SDL_GetGPUShaderFormats_REAL
#define SDL_GetGPUSwapchainTextureFormat SDL_GetGPUSwapchainTextureFormat_REAL
#define SDL_GetGamepadAppleSFSymbolsNameForAxis SDL_GetGamepadAppleSFSymbolsNameForAxis_REAL
#define SDL_GetGamepadAppleSFSymbolsNameForButton SDL_GetGamepadAppleSFSymbolsNameForButton_REAL
#define SDL_GetGamepadAxis SDL_GetGamepadAxis_REAL
#define SDL_GetGamepadAxisFromString SDL_GetGamepadAxisFromString_REAL
#define SDL_GetGamepadBindings SDL_GetGamepadBindings_REAL
#define SDL_GetGamepadButton SDL_GetGamepadButton_REAL
#define SDL_GetGamepadButtonFromString SDL_GetGamepadButtonFromString_REAL
#define SDL_GetGamepadButtonLabel SDL_GetGamepadButtonLabel_REAL
#define SDL_GetGamepadButtonLabelForType SDL_GetGamepadButtonLabelForType_REAL
#define SDL_GetGamepadConnectionState SDL_GetGamepadConnectionState_REAL
#define SDL_GetGamepadFirmwareVersion SDL_GetGamepadFirmwareVersion_REAL
#define SDL_GetGamepadFromID SDL_GetGamepadFromID_REAL
#define SDL_GetGamepadFromPlayerIndex SDL_GetGamepadFromPlayerIndex_REAL
#define SDL_GetGamepadGUIDForID SDL_GetGamepadGUIDForID_REAL
#define SDL_GetGamepadID SDL_GetGamepadID_REAL
#define SDL_GetGamepadJoystick SDL_GetGamepadJoystick_REAL
#define SDL_GetGamepadMapping SDL_GetGamepadMapping_REAL
#define SDL_GetGamepadMappingForGUID SDL_GetGamepadMappingForGUID_REAL
#define SDL_GetGamepadMappingForID SDL_GetGamepadMappingForID_REAL
#define SDL_GetGamepadMappings SDL_GetGamepadMappings_REAL
#define SDL_GetGamepadName SDL_GetGamepadName_REAL
#define SDL_GetGamepadNameForID SDL_GetGamepadNameForID_REAL
#define SDL_GetGamepadPath SDL_GetGamepadPath_REAL
#define SDL_GetGamepadPathForID SDL_GetGamepadPathForID_REAL
#define SDL_GetGamepadPlayerIndex SDL_GetGamepadPlayerIndex_REAL
#define SDL_GetGamepadPlayerIndexForID SDL_GetGamepadPlayerIndexForID_REAL
#define SDL_GetGamepadPowerInfo SDL_GetGamepadPowerInfo_REAL
#define SDL_GetGamepadProduct SDL_GetGamepadProduct_REAL
#define SDL_GetGamepadProductForID SDL_GetGamepadProductForID_REAL
#define SDL_GetGamepadProductVersion SDL_GetGamepadProductVersion_REAL
#define SDL_GetGamepadProductVersionForID SDL_GetGamepadProductVersionForID_REAL
#define SDL_GetGamepadProperties SDL_GetGamepadProperties_REAL
#define SDL_GetGamepadSensorData SDL_GetGamepadSensorData_REAL
#define SDL_GetGamepadSensorDataRate SDL_GetGamepadSensorDataRate_REAL
#define SDL_GetGamepadSerial SDL_GetGamepadSerial_REAL
#define SDL_GetGamepadSteamHandle SDL_GetGamepadSteamHandle_REAL
#define SDL_GetGamepadStringForAxis SDL_GetGamepadStringForAxis_REAL
#define SDL_GetGamepadStringForButton SDL_GetGamepadStringForButton_REAL
#define SDL_GetGamepadStringForType SDL_GetGamepadStringForType_REAL
#define SDL_GetGamepadTouchpadFinger SDL_GetGamepadTouchpadFinger_REAL
#define SDL_GetGamepadType SDL_GetGamepadType_REAL
#define SDL_GetGamepadTypeForID SDL_GetGamepadTypeForID_REAL
#define SDL_GetGamepadTypeFromString SDL_GetGamepadTypeFromString_REAL
#define SDL_GetGamepadVendor SDL_GetGamepadVendor_REAL
#define SDL_GetGamepadVendorForID SDL_GetGamepadVendorForID_REAL
#define SDL_GetGamepads SDL_GetGamepads_REAL
#define SDL_GetGlobalMouseState SDL_GetGlobalMouseState_REAL
#define SDL_GetGlobalProperties SDL_GetGlobalProperties_REAL
#define SDL_GetGrabbedWindow SDL_GetGrabbedWindow_REAL
#define SDL_GetHapticEffectStatus SDL_GetHapticEffectStatus_REAL
#define SDL_GetHapticFeatures SDL_GetHapticFeatures_REAL
#define SDL_GetHapticFromID SDL_GetHapticFromID_REAL
#define SDL_GetHapticID SDL_GetHapticID_REAL
#define SDL_GetHapticName SDL_GetHapticName_REAL
#define SDL_GetHapticNameForID SDL_GetHapticNameForID_REAL
#define SDL_GetHaptics SDL_GetHaptics_REAL
#define SDL_GetHint SDL_GetHint_REAL
#define SDL_GetHintBoolean SDL_GetHintBoolean_REAL
#define SDL_GetIOProperties SDL_GetIOProperties_REAL
#define SDL_GetIOSize SDL_GetIOSize_REAL
#define SDL_GetIOStatus SDL_GetIOStatus_REAL
#define SDL_GetJoystickAxis SDL_GetJoystickAxis_REAL
#define SDL_GetJoystickAxisInitialState SDL_GetJoystickAxisInitialState_REAL
#define SDL_GetJoystickBall SDL_GetJoystickBall_REAL
#define SDL_GetJoystickButton SDL_GetJoystickButton_REAL
#define SDL_GetJoystickConnectionState SDL_GetJoystickConnectionState_REAL
#define SDL_GetJoystickFirmwareVersion SDL_GetJoystickFirmwareVersion_REAL
#define SDL_GetJoystickFromID SDL_GetJoystickFromID_REAL
#define SDL_GetJoystickFromPlayerIndex SDL_GetJoystickFromPlayerIndex_REAL
#define SDL_GetJoystickGUID SDL_GetJoystickGUID_REAL
#define SDL_GetJoystickGUIDForID SDL_GetJoystickGUIDForID_REAL
#define SDL_GetJoystickGUIDInfo SDL_GetJoystickGUIDInfo_REAL
#define SDL_GetJoystickHat SDL_GetJoystickHat_REAL
#define SDL_GetJoystickID SDL_GetJoystickID_REAL
#define SDL_GetJoystickName SDL_GetJoystickName_REAL
#define SDL_GetJoystickNameForID SDL_GetJoystickNameForID_REAL
#define SDL_GetJoystickPath SDL_GetJoystickPath_REAL
#define SDL_GetJoystickPathForID SDL_GetJoystickPathForID_REAL
#define SDL_GetJoystickPlayerIndex SDL_GetJoystickPlayerIndex_REAL
#define SDL_GetJoystickPlayerIndexForID SDL_GetJoystickPlayerIndexForID_REAL
#define SDL_GetJoystickPowerInfo SDL_GetJoystickPowerInfo_REAL
#define SDL_GetJoystickProduct SDL_GetJoystickProduct_REAL
#define SDL_GetJoystickProductForID SDL_GetJoystickProductForID_REAL
#define SDL_GetJoystickProductVersion SDL_GetJoystickProductVersion_REAL
#define SDL_GetJoystickProductVersionForID SDL_GetJoystickProductVersionForID_REAL
#define SDL_GetJoystickProperties SDL_GetJoystickProperties_REAL
#define SDL_GetJoystickSerial SDL_GetJoystickSerial_REAL
#define SDL_GetJoystickType SDL_GetJoystickType_REAL
#define SDL_GetJoystickTypeForID SDL_GetJoystickTypeForID_REAL
#define SDL_GetJoystickVendor SDL_GetJoystickVendor_REAL
#define SDL_GetJoystickVendorForID SDL_GetJoystickVendorForID_REAL
#define SDL_GetJoysticks SDL_GetJoysticks_REAL
#define SDL_GetKeyFromName SDL_GetKeyFromName_REAL
#define SDL_GetKeyFromScancode SDL_GetKeyFromScancode_REAL
#define SDL_GetKeyName SDL_GetKeyName_REAL
#define SDL_GetKeyboardFocus SDL_GetKeyboardFocus_REAL
#define SDL_GetKeyboardNameForID SDL_GetKeyboardNameForID_REAL
#define SDL_GetKeyboardState SDL_GetKeyboardState_REAL
#define SDL_GetKeyboards SDL_GetKeyboards_REAL
#define SDL_GetLogOutputFunction SDL_GetLogOutputFunction_REAL
#define SDL_GetLogPriority SDL_GetLogPriority_REAL
#define SDL_GetMasksForPixelFormat SDL_GetMasksForPixelFormat_REAL
#define SDL_GetMaxHapticEffects SDL_GetMaxHapticEffects_REAL
#define SDL_GetMaxHapticEffectsPlaying SDL_GetMaxHapticEffectsPlaying_REAL
#define SDL_GetMemoryFunctions SDL_GetMemoryFunctions_REAL
#define SDL_GetMice SDL_GetMice_REAL
#define SDL_GetModState SDL_GetModState_REAL
#define SDL_GetMouseFocus SDL_GetMouseFocus_REAL
#define SDL_GetMouseNameForID SDL_GetMouseNameForID_REAL
#define SDL_GetMouseState SDL_GetMouseState_REAL
#define SDL_GetNaturalDisplayOrientation SDL_GetNaturalDisplayOrientation_REAL
#define SDL_GetNumAllocations SDL_GetNumAllocations_REAL
#define SDL_GetNumAudioDrivers SDL_GetNumAudioDrivers_REAL
#define SDL_GetNumCameraDrivers SDL_GetNumCameraDrivers_REAL
#define SDL_GetNumGPUDrivers SDL_GetNumGPUDrivers_REAL
#define SDL_GetNumGamepadTouchpadFingers SDL_GetNumGamepadTouchpadFingers_REAL
#define SDL_GetNumGamepadTouchpads SDL_GetNumGamepadTouchpads_REAL
#define SDL_GetNumHapticAxes SDL_GetNumHapticAxes_REAL
#define SDL_GetNumJoystickAxes SDL_GetNumJoystickAxes_REAL
#define SDL_GetNumJoystickBalls SDL_GetNumJoystickBalls_REAL
#define SDL_GetNumJoystickButtons SDL_GetNumJoystickButtons_REAL
#define SDL_GetNumJoystickHats SDL_GetNumJoystickHats_REAL
#define SDL_GetNumLogicalCPUCores SDL_GetNumLogicalCPUCores_REAL
#define SDL_GetNumRenderDrivers SDL_GetNumRenderDrivers_REAL
#define SDL_GetNumVideoDrivers SDL_GetNumVideoDrivers_REAL
#define SDL_GetNumberProperty SDL_GetNumberProperty_REAL
#define SDL_GetOriginalMemoryFunctions SDL_GetOriginalMemoryFunctions_REAL
#define SDL_GetPathInfo SDL_GetPathInfo_REAL
#define SDL_GetPerformanceCounter SDL_GetPerformanceCounter_REAL
#define SDL_GetPerformanceFrequency SDL_GetPerformanceFrequency_REAL
#define SDL_GetPixelFormatDetails SDL_GetPixelFormatDetails_REAL
#define SDL_GetPixelFormatForMasks SDL_GetPixelFormatForMasks_REAL
#define SDL_GetPixelFormatName SDL_GetPixelFormatName_REAL
#define SDL_GetPlatform SDL_GetPlatform_REAL
#define SDL_GetPointerProperty SDL_GetPointerProperty_REAL
#define SDL_GetPowerInfo SDL_GetPowerInfo_REAL
#define SDL_GetPrefPath SDL_GetPrefPath_REAL
#define SDL_GetPreferredLocales SDL_GetPreferredLocales_REAL
#define SDL_GetPrimaryDisplay SDL_GetPrimaryDisplay_REAL
#define SDL_GetPrimarySelectionText SDL_GetPrimarySelectionText_REAL
#define SDL_GetProcessInput SDL_GetProcessInput_REAL
#define SDL_GetProcessOutput SDL_GetProcessOutput_REAL
#define SDL_GetProcessProperties SDL_GetProcessProperties_REAL
#define SDL_GetPropertyType SDL_GetPropertyType_REAL
#define SDL_GetRGB SDL_GetRGB_REAL
#define SDL_GetRGBA SDL_GetRGBA_REAL
#define SDL_GetRealGamepadType SDL_GetRealGamepadType_REAL
#define SDL_GetRealGamepadTypeForID SDL_GetRealGamepadTypeForID_REAL
#define SDL_GetRectAndLineIntersection SDL_GetRectAndLineIntersection_REAL
#define SDL_GetRectAndLineIntersectionFloat SDL_GetRectAndLineIntersectionFloat_REAL
#define SDL_GetRectEnclosingPoints SDL_GetRectEnclosingPoints_REAL
#define SDL_GetRectEnclosingPointsFloat SDL_GetRectEnclosingPointsFloat_REAL
#define SDL_GetRectIntersection SDL_GetRectIntersection_REAL
#define SDL_GetRectIntersectionFloat SDL_GetRectIntersectionFloat_REAL
#define SDL_GetRectUnion SDL_GetRectUnion_REAL
#define SDL_GetRectUnionFloat SDL_GetRectUnionFloat_REAL
#define SDL_GetRelativeMouseState SDL_GetRelativeMouseState_REAL
#define SDL_GetRenderClipRect SDL_GetRenderClipRect_REAL
#define SDL_GetRenderColorScale SDL_GetRenderColorScale_REAL
#define SDL_GetRenderDrawBlendMode SDL_GetRenderDrawBlendMode_REAL
#define SDL_GetRenderDrawColor SDL_GetRenderDrawColor_REAL
#define SDL_GetRenderDrawColorFloat SDL_GetRenderDrawColorFloat_REAL
#define SDL_GetRenderDriver SDL_GetRenderDriver_REAL
#define SDL_GetRenderLogicalPresentation SDL_GetRenderLogicalPresentation_REAL
#define SDL_GetRenderLogicalPresentationRect SDL_GetRenderLogicalPresentationRect_REAL
#define SDL_GetRenderMetalCommandEncoder SDL_GetRenderMetalCommandEncoder_REAL
#define SDL_GetRenderMetalLayer SDL_GetRenderMetalLayer_REAL
#define SDL_GetRenderOutputSize SDL_GetRenderOutputSize_REAL
#define SDL_GetRenderSafeArea SDL_GetRenderSafeArea_REAL
#define SDL_GetRenderScale SDL_GetRenderScale_REAL
#define SDL_GetRenderTarget SDL_GetRenderTarget_REAL
#define SDL_GetRenderVSync SDL_GetRenderVSync_REAL
#define SDL_GetRenderViewport SDL_GetRenderViewport_REAL
#define SDL_GetRenderWindow SDL_GetRenderWindow_REAL
#define SDL_GetRenderer SDL_GetRenderer_REAL
#define SDL_GetRendererFromTexture SDL_GetRendererFromTexture_REAL
#define SDL_GetRendererName SDL_GetRendererName_REAL
#define SDL_GetRendererProperties SDL_GetRendererProperties_REAL
#define SDL_GetRevision SDL_GetRevision_REAL
#define SDL_GetSIMDAlignment SDL_GetSIMDAlignment_REAL
#define SDL_GetScancodeFromKey SDL_GetScancodeFromKey_REAL
#define SDL_GetScancodeFromName SDL_GetScancodeFromName_REAL
#define SDL_GetScancodeName SDL_GetScancodeName_REAL
#define SDL_GetSemaphoreValue SDL_GetSemaphoreValue_REAL
#define SDL_GetSensorData SDL_GetSensorData_REAL
#define SDL_GetSensorFromID SDL_GetSensorFromID_REAL
#define SDL_GetSensorID SDL_GetSensorID_REAL
#define SDL_GetSensorName SDL_GetSensorName_REAL
#define SDL_GetSensorNameForID SDL_GetSensorNameForID_REAL
#define SDL_GetSensorNonPortableType SDL_GetSensorNonPortableType_REAL
#define SDL_GetSensorNonPortableTypeForID SDL_GetSensorNonPortableTypeForID_REAL
#define SDL_GetSensorProperties SDL_GetSensorProperties_REAL
#define SDL_GetSensorType SDL_GetSensorType_REAL
#define SDL_GetSensorTypeForID SDL_GetSensorTypeForID_REAL
#define SDL_GetSensors SDL_GetSensors_REAL
#define SDL_GetSilenceValueForFormat SDL_GetSilenceValueForFormat_REAL
#define SDL_GetStorageFileSize SDL_GetStorageFileSize_REAL
#define SDL_GetStoragePathInfo SDL_GetStoragePathInfo_REAL
#define SDL_GetStorageSpaceRemaining SDL_GetStorageSpaceRemaining_REAL
#define SDL_GetStringProperty SDL_GetStringProperty_REAL
#define SDL_GetSurfaceAlphaMod SDL_GetSurfaceAlphaMod_REAL
#define SDL_GetSurfaceBlendMode SDL_GetSurfaceBlendMode_REAL
#define SDL_GetSurfaceClipRect SDL_GetSurfaceClipRect_REAL
#define SDL_GetSurfaceColorKey SDL_GetSurfaceColorKey_REAL
#define SDL_GetSurfaceColorMod SDL_GetSurfaceColorMod_REAL
#define SDL_GetSurfaceColorspace SDL_GetSurfaceColorspace_REAL
#define SDL_GetSurfaceImages SDL_GetSurfaceImages_REAL
#define SDL_GetSurfacePalette SDL_GetSurfacePalette_REAL
#define SDL_GetSurfaceProperties SDL_GetSurfaceProperties_REAL
#define SDL_GetSystemRAM SDL_GetSystemRAM_REAL
#define SDL_GetSystemTheme SDL_GetSystemTheme_REAL
#define SDL_GetTLS SDL_GetTLS_REAL
#define SDL_GetTextInputArea SDL_GetTextInputArea_REAL
#define SDL_GetTextureAlphaMod SDL_GetTextureAlphaMod_REAL
#define SDL_GetTextureAlphaModFloat SDL_GetTextureAlphaModFloat_REAL
#define SDL_GetTextureBlendMode SDL_GetTextureBlendMode_REAL
#define SDL_GetTextureColorMod SDL_GetTextureColorMod_REAL
#define SDL_GetTextureColorModFloat SDL_GetTextureColorModFloat_REAL
#define SDL_GetTextureProperties SDL_GetTextureProperties_REAL
#define SDL_GetTextureScaleMode SDL_GetTextureScaleMode_REAL
#define SDL_GetTextureSize SDL_GetTextureSize_REAL
#define SDL_GetThreadID SDL_GetThreadID_REAL
#define SDL_GetThreadName SDL_GetThreadName_REAL
#define SDL_GetTicks SDL_GetTicks_REAL
#define SDL_GetTicksNS SDL_GetTicksNS_REAL
#define SDL_GetTouchDeviceName SDL_GetTouchDeviceName_REAL
#define SDL_GetTouchDeviceType SDL_GetTouchDeviceType_REAL
#define SDL_GetTouchDevices SDL_GetTouchDevices_REAL
#define SDL_GetTouchFingers SDL_GetTouchFingers_REAL
#define SDL_GetUserFolder SDL_GetUserFolder_REAL
#define SDL_GetVersion SDL_GetVersion_REAL
#define SDL_GetVideoDriver SDL_GetVideoDriver_REAL
#define SDL_GetWindowAspectRatio SDL_GetWindowAspectRatio_REAL
#define SDL_GetWindowBordersSize SDL_GetWindowBordersSize_REAL
#define SDL_GetWindowDisplayScale SDL_GetWindowDisplayScale_REAL
#define SDL_GetWindowFlags SDL_GetWindowFlags_REAL
#define SDL_GetWindowFromEvent SDL_GetWindowFromEvent_REAL
#define SDL_GetWindowFromID SDL_GetWindowFromID_REAL
#define SDL_GetWindowFullscreenMode SDL_GetWindowFullscreenMode_REAL
#define SDL_GetWindowICCProfile SDL_GetWindowICCProfile_REAL
#define SDL_GetWindowID SDL_GetWindowID_REAL
#define SDL_GetWindowKeyboardGrab SDL_GetWindowKeyboardGrab_REAL
#define SDL_GetWindowMaximumSize SDL_GetWindowMaximumSize_REAL
#define SDL_GetWindowMinimumSize SDL_GetWindowMinimumSize_REAL
#define SDL_GetWindowMouseGrab SDL_GetWindowMouseGrab_REAL
#define SDL_GetWindowMouseRect SDL_GetWindowMouseRect_REAL
#define SDL_GetWindowOpacity SDL_GetWindowOpacity_REAL
#define SDL_GetWindowParent SDL_GetWindowParent_REAL
#define SDL_GetWindowPixelDensity SDL_GetWindowPixelDensity_REAL
#define SDL_GetWindowPixelFormat SDL_GetWindowPixelFormat_REAL
#define SDL_GetWindowPosition SDL_GetWindowPosition_REAL
#define SDL_GetWindowProperties SDL_GetWindowProperties_REAL
#define SDL_GetWindowRelativeMouseMode SDL_GetWindowRelativeMouseMode_REAL
#define SDL_GetWindowSafeArea SDL_GetWindowSafeArea_REAL
#define SDL_GetWindowSize SDL_GetWindowSize_REAL
#define SDL_GetWindowSizeInPixels SDL_GetWindowSizeInPixels_REAL
#define SDL_GetWindowSurface SDL_GetWindowSurface_REAL
#define SDL_GetWindowSurfaceVSync SDL_GetWindowSurfaceVSync_REAL
#define SDL_GetWindowTitle SDL_GetWindowTitle_REAL
#define SDL_GetWindows SDL_GetWindows_REAL
#define SDL_GlobDirectory SDL_GlobDirectory_REAL
#define SDL_GlobStorageDirectory SDL_GlobStorageDirectory_REAL
#define SDL_HapticEffectSupported SDL_HapticEffectSupported_REAL
#define SDL_HapticRumbleSupported SDL_HapticRumbleSupported_REAL
#define SDL_HasARMSIMD SDL_HasARMSIMD_REAL
#define SDL_HasAVX SDL_HasAVX_REAL
#define SDL_HasAVX2 SDL_HasAVX2_REAL
#define SDL_HasAVX512F SDL_HasAVX512F_REAL
#define SDL_HasAltiVec SDL_HasAltiVec_REAL
#define SDL_HasClipboardData SDL_HasClipboardData_REAL
#define SDL_HasClipboardText SDL_HasClipboardText_REAL
#define SDL_HasEvent SDL_HasEvent_REAL
#define SDL_HasEvents SDL_HasEvents_REAL
#define SDL_HasGamepad SDL_HasGamepad_REAL
#define SDL_HasJoystick SDL_HasJoystick_REAL
#define SDL_HasKeyboard SDL_HasKeyboard_REAL
#define SDL_HasLASX SDL_HasLASX_REAL
#define SDL_HasLSX SDL_HasLSX_REAL
#define SDL_HasMMX SDL_HasMMX_REAL
#define SDL_HasMouse SDL_HasMouse_REAL
#define SDL_HasNEON SDL_HasNEON_REAL
#define SDL_HasPrimarySelectionText SDL_HasPrimarySelectionText_REAL
#define SDL_HasProperty SDL_HasProperty_REAL
#define SDL_HasRectIntersection SDL_HasRectIntersection_REAL
#define SDL_HasRectIntersectionFloat SDL_HasRectIntersectionFloat_REAL
#define SDL_HasSSE SDL_HasSSE_REAL
#define SDL_HasSSE2 SDL_HasSSE2_REAL
#define SDL_HasSSE3 SDL_HasSSE3_REAL
#define SDL_HasSSE41 SDL_HasSSE41_REAL
#define SDL_HasSSE42 SDL_HasSSE42_REAL
#define SDL_HasScreenKeyboardSupport SDL_HasScreenKeyboardSupport_REAL
#define SDL_HideCursor SDL_HideCursor_REAL
#define SDL_HideWindow SDL_HideWindow_REAL
#define SDL_IOFromConstMem SDL_IOFromConstMem_REAL
#define SDL_IOFromDynamicMem SDL_IOFromDynamicMem_REAL
#define SDL_IOFromFile SDL_IOFromFile_REAL
#define SDL_IOFromMem SDL_IOFromMem_REAL
#define SDL_IOprintf SDL_IOprintf_REAL
#define SDL_IOvprintf SDL_IOvprintf_REAL
#define SDL_Init SDL_Init_REAL
#define SDL_InitHapticRumble SDL_InitHapticRumble_REAL
#define SDL_InitSubSystem SDL_InitSubSystem_REAL
#define SDL_InsertGPUDebugLabel SDL_InsertGPUDebugLabel_REAL
#define SDL_IsChromebook    SDL_IsChromebook_REAL
#define SDL_IsDeXMode   SDL_IsDeXMode_REAL
#define SDL_IsGamepad SDL_IsGamepad_REAL
#define SDL_IsJoystickHaptic SDL_IsJoystickHaptic_REAL
#define SDL_IsJoystickVirtual SDL_IsJoystickVirtual_REAL
#define SDL_IsMouseHaptic SDL_IsMouseHaptic_REAL
#define SDL_IsTV SDL_IsTV_REAL
#define SDL_IsTablet SDL_IsTablet_REAL
#define SDL_JoystickConnected SDL_JoystickConnected_REAL
#define SDL_JoystickEventsEnabled SDL_JoystickEventsEnabled_REAL
#define SDL_KillProcess SDL_KillProcess_REAL
#define SDL_LoadBMP SDL_LoadBMP_REAL
#define SDL_LoadBMP_IO SDL_LoadBMP_IO_REAL
#define SDL_LoadFile SDL_LoadFile_REAL
#define SDL_LoadFile_IO SDL_LoadFile_IO_REAL
#define SDL_LoadFunction SDL_LoadFunction_REAL
#define SDL_LoadObject SDL_LoadObject_REAL
#define SDL_LoadWAV SDL_LoadWAV_REAL
#define SDL_LoadWAV_IO SDL_LoadWAV_IO_REAL
#define SDL_LockAudioStream SDL_LockAudioStream_REAL
#define SDL_LockJoysticks SDL_LockJoysticks_REAL
#define SDL_LockMutex SDL_LockMutex_REAL
#define SDL_LockProperties SDL_LockProperties_REAL
#define SDL_LockRWLockForReading SDL_LockRWLockForReading_REAL
#define SDL_LockRWLockForWriting SDL_LockRWLockForWriting_REAL
#define SDL_LockSpinlock SDL_LockSpinlock_REAL
#define SDL_LockSurface SDL_LockSurface_REAL
#define SDL_LockTexture SDL_LockTexture_REAL
#define SDL_LockTextureToSurface SDL_LockTextureToSurface_REAL
#define SDL_Log SDL_Log_REAL
#define SDL_LogCritical SDL_LogCritical_REAL
#define SDL_LogDebug    SDL_LogDebug_REAL
#define SDL_LogError    SDL_LogError_REAL
#define SDL_LogInfo SDL_LogInfo_REAL
#define SDL_LogMessage  SDL_LogMessage_REAL
#define SDL_LogMessageV SDL_LogMessageV_REAL
#define SDL_LogTrace SDL_LogTrace_REAL
#define SDL_LogVerbose  SDL_LogVerbose_REAL
#define SDL_LogWarn SDL_LogWarn_REAL
#define SDL_MapGPUTransferBuffer SDL_MapGPUTransferBuffer_REAL
#define SDL_MapRGB SDL_MapRGB_REAL
#define SDL_MapRGBA SDL_MapRGBA_REAL
#define SDL_MapSurfaceRGB SDL_MapSurfaceRGB_REAL
#define SDL_MapSurfaceRGBA SDL_MapSurfaceRGBA_REAL
#define SDL_MaximizeWindow SDL_MaximizeWindow_REAL
#define SDL_MemoryBarrierAcquireFunction SDL_MemoryBarrierAcquireFunction_REAL
#define SDL_MemoryBarrierReleaseFunction SDL_MemoryBarrierReleaseFunction_REAL
#define SDL_Metal_CreateView SDL_Metal_CreateView_REAL
#define SDL_Metal_DestroyView SDL_Metal_DestroyView_REAL
#define SDL_Metal_GetLayer SDL_Metal_GetLayer_REAL
#define SDL_MinimizeWindow SDL_MinimizeWindow_REAL
#define SDL_MixAudio SDL_MixAudio_REAL
#define SDL_OnApplicationDidChangeStatusBarOrientation SDL_OnApplicationDidChangeStatusBarOrientation_REAL
#define SDL_OnApplicationDidEnterBackground SDL_OnApplicationDidEnterBackground_REAL
#define SDL_OnApplicationDidEnterForeground SDL_OnApplicationDidEnterForeground_REAL
#define SDL_OnApplicationDidReceiveMemoryWarning SDL_OnApplicationDidReceiveMemoryWarning_REAL
#define SDL_OnApplicationWillEnterBackground SDL_OnApplicationWillEnterBackground_REAL
#define SDL_OnApplicationWillEnterForeground SDL_OnApplicationWillEnterForeground_REAL
#define SDL_OnApplicationWillTerminate SDL_OnApplicationWillTerminate_REAL
#define SDL_OpenAudioDevice SDL_OpenAudioDevice_REAL
#define SDL_OpenAudioDeviceStream SDL_OpenAudioDeviceStream_REAL
#define SDL_OpenCamera SDL_OpenCamera_REAL
#define SDL_OpenFileStorage SDL_OpenFileStorage_REAL
#define SDL_OpenGamepad SDL_OpenGamepad_REAL
#define SDL_OpenHaptic SDL_OpenHaptic_REAL
#define SDL_OpenHapticFromJoystick SDL_OpenHapticFromJoystick_REAL
#define SDL_OpenHapticFromMouse SDL_OpenHapticFromMouse_REAL
#define SDL_OpenIO SDL_OpenIO_REAL
#define SDL_OpenJoystick SDL_OpenJoystick_REAL
#define SDL_OpenSensor SDL_OpenSensor_REAL
#define SDL_OpenStorage SDL_OpenStorage_REAL
#define SDL_OpenTitleStorage SDL_OpenTitleStorage_REAL
#define SDL_OpenURL SDL_OpenURL_REAL
#define SDL_OpenUserStorage SDL_OpenUserStorage_REAL
#define SDL_OutOfMemory SDL_OutOfMemory_REAL
#define SDL_PauseAudioDevice SDL_PauseAudioDevice_REAL
#define SDL_PauseAudioStreamDevice SDL_PauseAudioStreamDevice_REAL
#define SDL_PauseHaptic SDL_PauseHaptic_REAL
#define SDL_PeepEvents SDL_PeepEvents_REAL
#define SDL_PlayHapticRumble SDL_PlayHapticRumble_REAL
#define SDL_PollEvent SDL_PollEvent_REAL
#define SDL_PopGPUDebugGroup SDL_PopGPUDebugGroup_REAL
#define SDL_PremultiplyAlpha SDL_PremultiplyAlpha_REAL
#define SDL_PremultiplySurfaceAlpha SDL_PremultiplySurfaceAlpha_REAL
#define SDL_PumpEvents SDL_PumpEvents_REAL
#define SDL_PushEvent SDL_PushEvent_REAL
#define SDL_PushGPUComputeUniformData SDL_PushGPUComputeUniformData_REAL
#define SDL_PushGPUDebugGroup SDL_PushGPUDebugGroup_REAL
#define SDL_PushGPUFragmentUniformData SDL_PushGPUFragmentUniformData_REAL
#define SDL_PushGPUVertexUniformData SDL_PushGPUVertexUniformData_REAL
#define SDL_PutAudioStreamData SDL_PutAudioStreamData_REAL
#define SDL_QueryGPUFence SDL_QueryGPUFence_REAL
#define SDL_Quit SDL_Quit_REAL
#define SDL_QuitSubSystem SDL_QuitSubSystem_REAL
#define SDL_RaiseWindow SDL_RaiseWindow_REAL
#define SDL_ReadIO SDL_ReadIO_REAL
#define SDL_ReadProcess SDL_ReadProcess_REAL
#define SDL_ReadS16BE SDL_ReadS16BE_REAL
#define SDL_ReadS16LE SDL_ReadS16LE_REAL
#define SDL_ReadS32BE SDL_ReadS32BE_REAL
#define SDL_ReadS32LE SDL_ReadS32LE_REAL
#define SDL_ReadS64BE SDL_ReadS64BE_REAL
#define SDL_ReadS64LE SDL_ReadS64LE_REAL
#define SDL_ReadS8 SDL_ReadS8_REAL
#define SDL_ReadStorageFile SDL_ReadStorageFile_REAL
#define SDL_ReadSurfacePixel SDL_ReadSurfacePixel_REAL
#define SDL_ReadSurfacePixelFloat SDL_ReadSurfacePixelFloat_REAL
#define SDL_ReadU16BE SDL_ReadU16BE_REAL
#define SDL_ReadU16LE SDL_ReadU16LE_REAL
#define SDL_ReadU32BE SDL_ReadU32BE_REAL
#define SDL_ReadU32LE SDL_ReadU32LE_REAL
#define SDL_ReadU64BE SDL_ReadU64BE_REAL
#define SDL_ReadU64LE SDL_ReadU64LE_REAL
#define SDL_ReadU8 SDL_ReadU8_REAL
#define SDL_RegisterApp SDL_RegisterApp_REAL
#define SDL_RegisterEvents SDL_RegisterEvents_REAL
#define SDL_ReleaseCameraFrame SDL_ReleaseCameraFrame_REAL
#define SDL_ReleaseGPUBuffer SDL_ReleaseGPUBuffer_REAL
#define SDL_ReleaseGPUComputePipeline SDL_ReleaseGPUComputePipeline_REAL
#define SDL_ReleaseGPUFence SDL_ReleaseGPUFence_REAL
#define SDL_ReleaseGPUGraphicsPipeline SDL_ReleaseGPUGraphicsPipeline_REAL
#define SDL_ReleaseGPUSampler SDL_ReleaseGPUSampler_REAL
#define SDL_ReleaseGPUShader SDL_ReleaseGPUShader_REAL
#define SDL_ReleaseGPUTexture SDL_ReleaseGPUTexture_REAL
#define SDL_ReleaseGPUTransferBuffer SDL_ReleaseGPUTransferBuffer_REAL
#define SDL_ReleaseWindowFromGPUDevice SDL_ReleaseWindowFromGPUDevice_REAL
#define SDL_ReloadGamepadMappings SDL_ReloadGamepadMappings_REAL
#define SDL_RemoveEventWatch SDL_RemoveEventWatch_REAL
#define SDL_RemoveHintCallback SDL_RemoveHintCallback_REAL
#define SDL_RemovePath SDL_RemovePath_REAL
#define SDL_RemoveStoragePath SDL_RemoveStoragePath_REAL
#define SDL_RemoveSurfaceAlternateImages SDL_RemoveSurfaceAlternateImages_REAL
#define SDL_RemoveTimer SDL_RemoveTimer_REAL
#define SDL_RenamePath SDL_RenamePath_REAL
#define SDL_RenameStoragePath SDL_RenameStoragePath_REAL
#define SDL_RenderClear SDL_RenderClear_REAL
#define SDL_RenderClipEnabled SDL_RenderClipEnabled_REAL
#define SDL_RenderCoordinatesFromWindow SDL_RenderCoordinatesFromWindow_REAL
#define SDL_RenderCoordinatesToWindow SDL_RenderCoordinatesToWindow_REAL
#define SDL_RenderFillRect SDL_RenderFillRect_REAL
#define SDL_RenderFillRects SDL_RenderFillRects_REAL
#define SDL_RenderGeometry SDL_RenderGeometry_REAL
#define SDL_RenderGeometryRaw SDL_RenderGeometryRaw_REAL
#define SDL_RenderLine SDL_RenderLine_REAL
#define SDL_RenderLines SDL_RenderLines_REAL
#define SDL_RenderPoint SDL_RenderPoint_REAL
#define SDL_RenderPoints SDL_RenderPoints_REAL
#define SDL_RenderPresent SDL_RenderPresent_REAL
#define SDL_RenderReadPixels SDL_RenderReadPixels_REAL
#define SDL_RenderRect SDL_RenderRect_REAL
#define SDL_RenderRects SDL_RenderRects_REAL
#define SDL_RenderTexture SDL_RenderTexture_REAL
#define SDL_RenderTexture9Grid SDL_RenderTexture9Grid_REAL
#define SDL_RenderTextureRotated SDL_RenderTextureRotated_REAL
#define SDL_RenderTextureTiled SDL_RenderTextureTiled_REAL
#define SDL_RenderViewportSet SDL_RenderViewportSet_REAL
#define SDL_ReportAssertion SDL_ReportAssertion_REAL
#define SDL_RequestAndroidPermission SDL_RequestAndroidPermission_REAL
#define SDL_ResetAssertionReport SDL_ResetAssertionReport_REAL
#define SDL_ResetHint SDL_ResetHint_REAL
#define SDL_ResetHints SDL_ResetHints_REAL
#define SDL_ResetKeyboard SDL_ResetKeyboard_REAL
#define SDL_ResetLogPriorities SDL_ResetLogPriorities_REAL
#define SDL_RestoreWindow SDL_RestoreWindow_REAL
#define SDL_ResumeAudioDevice SDL_ResumeAudioDevice_REAL
#define SDL_ResumeAudioStreamDevice SDL_ResumeAudioStreamDevice_REAL
#define SDL_ResumeHaptic SDL_ResumeHaptic_REAL
#define SDL_RumbleGamepad SDL_RumbleGamepad_REAL
#define SDL_RumbleGamepadTriggers SDL_RumbleGamepadTriggers_REAL
#define SDL_RumbleJoystick SDL_RumbleJoystick_REAL
#define SDL_RumbleJoystickTriggers SDL_RumbleJoystickTriggers_REAL
#define SDL_RunApp SDL_RunApp_REAL
#define SDL_RunHapticEffect SDL_RunHapticEffect_REAL
#define SDL_SaveBMP SDL_SaveBMP_REAL
#define SDL_SaveBMP_IO SDL_SaveBMP_IO_REAL
#define SDL_ScaleSurface SDL_ScaleSurface_REAL
#define SDL_ScreenKeyboardShown SDL_ScreenKeyboardShown_REAL
#define SDL_ScreenSaverEnabled SDL_ScreenSaverEnabled_REAL
#define SDL_SeekIO SDL_SeekIO_REAL
#define SDL_SendAndroidBackButton SDL_SendAndroidBackButton_REAL
#define SDL_SendAndroidMessage SDL_SendAndroidMessage_REAL
#define SDL_SendGamepadEffect SDL_SendGamepadEffect_REAL
#define SDL_SendJoystickEffect SDL_SendJoystickEffect_REAL
#define SDL_SendJoystickVirtualSensorData SDL_SendJoystickVirtualSensorData_REAL
#define SDL_SetAppMetadata SDL_SetAppMetadata_REAL
#define SDL_SetAppMetadataProperty SDL_SetAppMetadataProperty_REAL
#define SDL_SetAssertionHandler SDL_SetAssertionHandler_REAL
#define SDL_SetAtomicInt SDL_SetAtomicInt_REAL
#define SDL_SetAtomicPointer SDL_SetAtomicPointer_REAL
#define SDL_SetAtomicU32 SDL_SetAtomicU32_REAL
#define SDL_SetAudioDeviceGain SDL_SetAudioDeviceGain_REAL
#define SDL_SetAudioPostmixCallback SDL_SetAudioPostmixCallback_REAL
#define SDL_SetAudioStreamFormat SDL_SetAudioStreamFormat_REAL
#define SDL_SetAudioStreamFrequencyRatio SDL_SetAudioStreamFrequencyRatio_REAL
#define SDL_SetAudioStreamGain SDL_SetAudioStreamGain_REAL
#define SDL_SetAudioStreamGetCallback SDL_SetAudioStreamGetCallback_REAL
#define SDL_SetAudioStreamInputChannelMap SDL_SetAudioStreamInputChannelMap_REAL
#define SDL_SetAudioStreamOutputChannelMap SDL_SetAudioStreamOutputChannelMap_REAL
#define SDL_SetAudioStreamPutCallback SDL_SetAudioStreamPutCallback_REAL
#define SDL_SetBooleanProperty SDL_SetBooleanProperty_REAL
#define SDL_SetClipboardData SDL_SetClipboardData_REAL
#define SDL_SetClipboardText SDL_SetClipboardText_REAL
#define SDL_SetCurrentThreadPriority SDL_SetCurrentThreadPriority_REAL
#define SDL_SetCursor SDL_SetCursor_REAL
#define SDL_SetEnvironmentVariable SDL_SetEnvironmentVariable_REAL
#define SDL_SetError    SDL_SetError_REAL
#define SDL_SetEventEnabled SDL_SetEventEnabled_REAL
#define SDL_SetEventFilter SDL_SetEventFilter_REAL
#define SDL_SetFloatProperty SDL_SetFloatProperty_REAL
#define SDL_SetGPUBlendConstants SDL_SetGPUBlendConstants_REAL
#define SDL_SetGPUBufferName SDL_SetGPUBufferName_REAL
#define SDL_SetGPUScissor SDL_SetGPUScissor_REAL
#define SDL_SetGPUStencilReference SDL_SetGPUStencilReference_REAL
#define SDL_SetGPUSwapchainParameters SDL_SetGPUSwapchainParameters_REAL
#define SDL_SetGPUTextureName SDL_SetGPUTextureName_REAL
#define SDL_SetGPUViewport SDL_SetGPUViewport_REAL
#define SDL_SetGamepadEventsEnabled SDL_SetGamepadEventsEnabled_REAL
#define SDL_SetGamepadLED SDL_SetGamepadLED_REAL
#define SDL_SetGamepadMapping SDL_SetGamepadMapping_REAL
#define SDL_SetGamepadPlayerIndex SDL_SetGamepadPlayerIndex_REAL
#define SDL_SetGamepadSensorEnabled SDL_SetGamepadSensorEnabled_REAL
#define SDL_SetHapticAutocenter SDL_SetHapticAutocenter_REAL
#define SDL_SetHapticGain SDL_SetHapticGain_REAL
#define SDL_SetHint SDL_SetHint_REAL
#define SDL_SetHintWithPriority SDL_SetHintWithPriority_REAL
#define SDL_SetInitialized SDL_SetInitialized_REAL
#define SDL_SetJoystickEventsEnabled SDL_SetJoystickEventsEnabled_REAL
#define SDL_SetJoystickLED SDL_SetJoystickLED_REAL
#define SDL_SetJoystickPlayerIndex SDL_SetJoystickPlayerIndex_REAL
#define SDL_SetJoystickVirtualAxis SDL_SetJoystickVirtualAxis_REAL
#define SDL_SetJoystickVirtualBall SDL_SetJoystickVirtualBall_REAL
#define SDL_SetJoystickVirtualButton SDL_SetJoystickVirtualButton_REAL
#define SDL_SetJoystickVirtualHat SDL_SetJoystickVirtualHat_REAL
#define SDL_SetJoystickVirtualTouchpad SDL_SetJoystickVirtualTouchpad_REAL
#define SDL_SetLinuxThreadPriority SDL_SetLinuxThreadPriority_REAL
#define SDL_SetLinuxThreadPriorityAndPolicy SDL_SetLinuxThreadPriorityAndPolicy_REAL
#define SDL_SetLogOutputFunction SDL_SetLogOutputFunction_REAL
#define SDL_SetLogPriorities SDL_SetLogPriorities_REAL
#define SDL_SetLogPriority SDL_SetLogPriority_REAL
#define SDL_SetLogPriorityPrefix SDL_SetLogPriorityPrefix_REAL
#define SDL_SetMainReady SDL_SetMainReady_REAL
#define SDL_SetMemoryFunctions SDL_SetMemoryFunctions_REAL
#define SDL_SetModState SDL_SetModState_REAL
#define SDL_SetNumberProperty SDL_SetNumberProperty_REAL
#define SDL_SetPaletteColors SDL_SetPaletteColors_REAL
#define SDL_SetPointerProperty SDL_SetPointerProperty_REAL
#define SDL_SetPointerPropertyWithCleanup SDL_SetPointerPropertyWithCleanup_REAL
#define SDL_SetPrimarySelectionText SDL_SetPrimarySelectionText_REAL
#define SDL_SetRenderClipRect SDL_SetRenderClipRect_REAL
#define SDL_SetRenderColorScale SDL_SetRenderColorScale_REAL
#define SDL_SetRenderDrawBlendMode SDL_SetRenderDrawBlendMode_REAL
#define SDL_SetRenderDrawColor SDL_SetRenderDrawColor_REAL
#define SDL_SetRenderDrawColorFloat SDL_SetRenderDrawColorFloat_REAL
#define SDL_SetRenderLogicalPresentation SDL_SetRenderLogicalPresentation_REAL
#define SDL_SetRenderScale SDL_SetRenderScale_REAL
#define SDL_SetRenderTarget SDL_SetRenderTarget_REAL
#define SDL_SetRenderVSync SDL_SetRenderVSync_REAL
#define SDL_SetRenderViewport SDL_SetRenderViewport_REAL
#define SDL_SetScancodeName SDL_SetScancodeName_REAL
#define SDL_SetStringProperty SDL_SetStringProperty_REAL
#define SDL_SetSurfaceAlphaMod SDL_SetSurfaceAlphaMod_REAL
#define SDL_SetSurfaceBlendMode SDL_SetSurfaceBlendMode_REAL
#define SDL_SetSurfaceClipRect SDL_SetSurfaceClipRect_REAL
#define SDL_SetSurfaceColorKey SDL_SetSurfaceColorKey_REAL
#define SDL_SetSurfaceColorMod SDL_SetSurfaceColorMod_REAL
#define SDL_SetSurfaceColorspace SDL_SetSurfaceColorspace_REAL
#define SDL_SetSurfacePalette SDL_SetSurfacePalette_REAL
#define SDL_SetSurfaceRLE SDL_SetSurfaceRLE_REAL
#define SDL_SetTLS SDL_SetTLS_REAL
#define SDL_SetTextInputArea SDL_SetTextInputArea_REAL
#define SDL_SetTextureAlphaMod SDL_SetTextureAlphaMod_REAL
#define SDL_SetTextureAlphaModFloat SDL_SetTextureAlphaModFloat_REAL
#define SDL_SetTextureBlendMode SDL_SetTextureBlendMode_REAL
#define SDL_SetTextureColorMod SDL_SetTextureColorMod_REAL
#define SDL_SetTextureColorModFloat SDL_SetTextureColorModFloat_REAL
#define SDL_SetTextureScaleMode SDL_SetTextureScaleMode_REAL
#define SDL_SetWindowAlwaysOnTop SDL_SetWindowAlwaysOnTop_REAL
#define SDL_SetWindowAspectRatio SDL_SetWindowAspectRatio_REAL
#define SDL_SetWindowBordered SDL_SetWindowBordered_REAL
#define SDL_SetWindowFocusable SDL_SetWindowFocusable_REAL
#define SDL_SetWindowFullscreen SDL_SetWindowFullscreen_REAL
#define SDL_SetWindowFullscreenMode SDL_SetWindowFullscreenMode_REAL
#define SDL_SetWindowHitTest SDL_SetWindowHitTest_REAL
#define SDL_SetWindowIcon SDL_SetWindowIcon_REAL
#define SDL_SetWindowKeyboardGrab SDL_SetWindowKeyboardGrab_REAL
#define SDL_SetWindowMaximumSize SDL_SetWindowMaximumSize_REAL
#define SDL_SetWindowMinimumSize SDL_SetWindowMinimumSize_REAL
#define SDL_SetWindowModal SDL_SetWindowModal_REAL
#define SDL_SetWindowMouseGrab SDL_SetWindowMouseGrab_REAL
#define SDL_SetWindowMouseRect SDL_SetWindowMouseRect_REAL
#define SDL_SetWindowOpacity SDL_SetWindowOpacity_REAL
#define SDL_SetWindowParent SDL_SetWindowParent_REAL
#define SDL_SetWindowPosition SDL_SetWindowPosition_REAL
#define SDL_SetWindowRelativeMouseMode SDL_SetWindowRelativeMouseMode_REAL
#define SDL_SetWindowResizable SDL_SetWindowResizable_REAL
#define SDL_SetWindowShape SDL_SetWindowShape_REAL
#define SDL_SetWindowSize SDL_SetWindowSize_REAL
#define SDL_SetWindowSurfaceVSync SDL_SetWindowSurfaceVSync_REAL
#define SDL_SetWindowTitle SDL_SetWindowTitle_REAL
#define SDL_SetWindowsMessageHook SDL_SetWindowsMessageHook_REAL
#define SDL_SetX11EventHook SDL_SetX11EventHook_REAL
#define SDL_SetiOSAnimationCallback SDL_SetiOSAnimationCallback_REAL
#define SDL_SetiOSEventPump SDL_SetiOSEventPump_REAL
#define SDL_ShouldInit SDL_ShouldInit_REAL
#define SDL_ShouldQuit SDL_ShouldQuit_REAL
#define SDL_ShowAndroidToast SDL_ShowAndroidToast_REAL
#define SDL_ShowCursor SDL_ShowCursor_REAL
#define SDL_ShowMessageBox SDL_ShowMessageBox_REAL
#define SDL_ShowOpenFileDialog SDL_ShowOpenFileDialog_REAL
#define SDL_ShowOpenFolderDialog SDL_ShowOpenFolderDialog_REAL
#define SDL_ShowSaveFileDialog SDL_ShowSaveFileDialog_REAL
#define SDL_ShowSimpleMessageBox SDL_ShowSimpleMessageBox_REAL
#define SDL_ShowWindow SDL_ShowWindow_REAL
#define SDL_ShowWindowSystemMenu SDL_ShowWindowSystemMenu_REAL
#define SDL_SignalCondition SDL_SignalCondition_REAL
#define SDL_SignalSemaphore SDL_SignalSemaphore_REAL
#define SDL_StartTextInput SDL_StartTextInput_REAL
#define SDL_StartTextInputWithProperties SDL_StartTextInputWithProperties_REAL
#define SDL_StepUTF8 SDL_StepUTF8_REAL
#define SDL_StopHapticEffect SDL_StopHapticEffect_REAL
#define SDL_StopHapticEffects SDL_StopHapticEffects_REAL
#define SDL_StopHapticRumble SDL_StopHapticRumble_REAL
#define SDL_StopTextInput SDL_StopTextInput_REAL
#define SDL_StorageReady SDL_StorageReady_REAL
#define SDL_StringToGUID SDL_StringToGUID_REAL
#define SDL_SubmitGPUCommandBuffer SDL_SubmitGPUCommandBuffer_REAL
#define SDL_SubmitGPUCommandBufferAndAcquireFence SDL_SubmitGPUCommandBufferAndAcquireFence_REAL
#define SDL_SurfaceHasAlternateImages SDL_SurfaceHasAlternateImages_REAL
#define SDL_SurfaceHasColorKey SDL_SurfaceHasColorKey_REAL
#define SDL_SurfaceHasRLE SDL_SurfaceHasRLE_REAL
#define SDL_SyncWindow SDL_SyncWindow_REAL
#define SDL_TellIO SDL_TellIO_REAL
#define SDL_TextInputActive SDL_TextInputActive_REAL
#define SDL_TimeFromWindows SDL_TimeFromWindows_REAL
#define SDL_TimeToDateTime SDL_TimeToDateTime_REAL
#define SDL_TimeToWindows SDL_TimeToWindows_REAL
#define SDL_TryLockMutex SDL_TryLockMutex_REAL
#define SDL_TryLockRWLockForReading SDL_TryLockRWLockForReading_REAL
#define SDL_TryLockRWLockForWriting SDL_TryLockRWLockForWriting_REAL
#define SDL_TryLockSpinlock SDL_TryLockSpinlock_REAL
#define SDL_TryWaitSemaphore SDL_TryWaitSemaphore_REAL
#define SDL_UCS4ToUTF8 SDL_UCS4ToUTF8_REAL
#define SDL_UnbindAudioStream SDL_UnbindAudioStream_REAL
#define SDL_UnbindAudioStreams SDL_UnbindAudioStreams_REAL
#define SDL_UnloadObject SDL_UnloadObject_REAL
#define SDL_UnlockAudioStream SDL_UnlockAudioStream_REAL
#define SDL_UnlockJoysticks SDL_UnlockJoysticks_REAL
#define SDL_UnlockMutex SDL_UnlockMutex_REAL
#define SDL_UnlockProperties SDL_UnlockProperties_REAL
#define SDL_UnlockRWLock SDL_UnlockRWLock_REAL
#define SDL_UnlockSpinlock SDL_UnlockSpinlock_REAL
#define SDL_UnlockSurface SDL_UnlockSurface_REAL
#define SDL_UnlockTexture SDL_UnlockTexture_REAL
#define SDL_UnmapGPUTransferBuffer SDL_UnmapGPUTransferBuffer_REAL
#define SDL_UnregisterApp SDL_UnregisterApp_REAL
#define SDL_UnsetEnvironmentVariable SDL_UnsetEnvironmentVariable_REAL
#define SDL_UpdateGamepads SDL_UpdateGamepads_REAL
#define SDL_UpdateHapticEffect SDL_UpdateHapticEffect_REAL
#define SDL_UpdateJoysticks SDL_UpdateJoysticks_REAL
#define SDL_UpdateNVTexture SDL_UpdateNVTexture_REAL
#define SDL_UpdateSensors SDL_UpdateSensors_REAL
#define SDL_UpdateTexture SDL_UpdateTexture_REAL
#define SDL_UpdateWindowSurface SDL_UpdateWindowSurface_REAL
#define SDL_UpdateWindowSurfaceRects SDL_UpdateWindowSurfaceRects_REAL
#define SDL_UpdateYUVTexture SDL_UpdateYUVTexture_REAL
#define SDL_UploadToGPUBuffer SDL_UploadToGPUBuffer_REAL
#define SDL_UploadToGPUTexture SDL_UploadToGPUTexture_REAL
#define SDL_Vulkan_CreateSurface SDL_Vulkan_CreateSurface_REAL
#define SDL_Vulkan_DestroySurface SDL_Vulkan_DestroySurface_REAL
#define SDL_Vulkan_GetInstanceExtensions SDL_Vulkan_GetInstanceExtensions_REAL
#define SDL_Vulkan_GetPresentationSupport SDL_Vulkan_GetPresentationSupport_REAL
#define SDL_Vulkan_GetVkGetInstanceProcAddr SDL_Vulkan_GetVkGetInstanceProcAddr_REAL
#define SDL_Vulkan_LoadLibrary SDL_Vulkan_LoadLibrary_REAL
#define SDL_Vulkan_UnloadLibrary SDL_Vulkan_UnloadLibrary_REAL
#define SDL_WaitCondition SDL_WaitCondition_REAL
#define SDL_WaitConditionTimeout SDL_WaitConditionTimeout_REAL
#define SDL_WaitEvent SDL_WaitEvent_REAL
#define SDL_WaitEventTimeout SDL_WaitEventTimeout_REAL
#define SDL_WaitForGPUFences SDL_WaitForGPUFences_REAL
#define SDL_WaitForGPUIdle SDL_WaitForGPUIdle_REAL
#define SDL_WaitProcess SDL_WaitProcess_REAL
#define SDL_WaitSemaphore SDL_WaitSemaphore_REAL
#define SDL_WaitSemaphoreTimeout SDL_WaitSemaphoreTimeout_REAL
#define SDL_WaitThread SDL_WaitThread_REAL
#define SDL_WarpMouseGlobal SDL_WarpMouseGlobal_REAL
#define SDL_WarpMouseInWindow SDL_WarpMouseInWindow_REAL
#define SDL_WasInit SDL_WasInit_REAL
#define SDL_WindowHasSurface SDL_WindowHasSurface_REAL
#define SDL_WindowSupportsGPUPresentMode SDL_WindowSupportsGPUPresentMode_REAL
#define SDL_WindowSupportsGPUSwapchainComposition SDL_WindowSupportsGPUSwapchainComposition_REAL
#define SDL_WriteIO SDL_WriteIO_REAL
#define SDL_WriteS16BE SDL_WriteS16BE_REAL
#define SDL_WriteS16LE SDL_WriteS16LE_REAL
#define SDL_WriteS32BE SDL_WriteS32BE_REAL
#define SDL_WriteS32LE SDL_WriteS32LE_REAL
#define SDL_WriteS64BE SDL_WriteS64BE_REAL
#define SDL_WriteS64LE SDL_WriteS64LE_REAL
#define SDL_WriteS8 SDL_WriteS8_REAL
#define SDL_WriteStorageFile SDL_WriteStorageFile_REAL
#define SDL_WriteSurfacePixel SDL_WriteSurfacePixel_REAL
#define SDL_WriteSurfacePixelFloat SDL_WriteSurfacePixelFloat_REAL
#define SDL_WriteU16BE SDL_WriteU16BE_REAL
#define SDL_WriteU16LE SDL_WriteU16LE_REAL
#define SDL_WriteU32BE SDL_WriteU32BE_REAL
#define SDL_WriteU32LE SDL_WriteU32LE_REAL
#define SDL_WriteU64BE SDL_WriteU64BE_REAL
#define SDL_WriteU64LE SDL_WriteU64LE_REAL
#define SDL_WriteU8 SDL_WriteU8_REAL
#define SDL_abs SDL_abs_REAL
#define SDL_acos SDL_acos_REAL
#define SDL_acosf SDL_acosf_REAL
#define SDL_aligned_alloc SDL_aligned_alloc_REAL
#define SDL_aligned_free SDL_aligned_free_REAL
#define SDL_asin SDL_asin_REAL
#define SDL_asinf SDL_asinf_REAL
#define SDL_asprintf SDL_asprintf_REAL
#define SDL_atan SDL_atan_REAL
#define SDL_atan2 SDL_atan2_REAL
#define SDL_atan2f SDL_atan2f_REAL
#define SDL_atanf SDL_atanf_REAL
#define SDL_atof SDL_atof_REAL
#define SDL_atoi SDL_atoi_REAL
#define SDL_bsearch SDL_bsearch_REAL
#define SDL_bsearch_r SDL_bsearch_r_REAL
#define SDL_calloc SDL_calloc_REAL
#define SDL_ceil SDL_ceil_REAL
#define SDL_ceilf SDL_ceilf_REAL
#define SDL_copysign SDL_copysign_REAL
#define SDL_copysignf SDL_copysignf_REAL
#define SDL_cos SDL_cos_REAL
#define SDL_cosf SDL_cosf_REAL
#define SDL_crc16 SDL_crc16_REAL
#define SDL_crc32 SDL_crc32_REAL
#define SDL_exp SDL_exp_REAL
#define SDL_expf SDL_expf_REAL
#define SDL_fabs SDL_fabs_REAL
#define SDL_fabsf SDL_fabsf_REAL
#define SDL_floor SDL_floor_REAL
#define SDL_floorf SDL_floorf_REAL
#define SDL_fmod SDL_fmod_REAL
#define SDL_fmodf SDL_fmodf_REAL
#define SDL_free SDL_free_REAL
#define SDL_getenv SDL_getenv_REAL
#define SDL_getenv_unsafe SDL_getenv_unsafe_REAL
#define SDL_hid_ble_scan SDL_hid_ble_scan_REAL
#define SDL_hid_close SDL_hid_close_REAL
#define SDL_hid_device_change_count SDL_hid_device_change_count_REAL
#define SDL_hid_enumerate SDL_hid_enumerate_REAL
#define SDL_hid_exit SDL_hid_exit_REAL
#define SDL_hid_free_enumeration SDL_hid_free_enumeration_REAL
#define SDL_hid_get_device_info SDL_hid_get_device_info_REAL
#define SDL_hid_get_feature_report SDL_hid_get_feature_report_REAL
#define SDL_hid_get_indexed_string SDL_hid_get_indexed_string_REAL
#define SDL_hid_get_input_report SDL_hid_get_input_report_REAL
#define SDL_hid_get_manufacturer_string SDL_hid_get_manufacturer_string_REAL
#define SDL_hid_get_product_string SDL_hid_get_product_string_REAL
#define SDL_hid_get_report_descriptor SDL_hid_get_report_descriptor_REAL
#define SDL_hid_get_serial_number_string SDL_hid_get_serial_number_string_REAL
#define SDL_hid_init SDL_hid_init_REAL
#define SDL_hid_open SDL_hid_open_REAL
#define SDL_hid_open_path SDL_hid_open_path_REAL
#define SDL_hid_read SDL_hid_read_REAL
#define SDL_hid_read_timeout SDL_hid_read_timeout_REAL
#define SDL_hid_send_feature_report SDL_hid_send_feature_report_REAL
#define SDL_hid_set_nonblocking SDL_hid_set_nonblocking_REAL
#define SDL_hid_write SDL_hid_write_REAL
#define SDL_iconv SDL_iconv_REAL
#define SDL_iconv_close SDL_iconv_close_REAL
#define SDL_iconv_open SDL_iconv_open_REAL
#define SDL_iconv_string SDL_iconv_string_REAL
#define SDL_isalnum SDL_isalnum_REAL
#define SDL_isalpha SDL_isalpha_REAL
#define SDL_isblank SDL_isblank_REAL
#define SDL_iscntrl SDL_iscntrl_REAL
#define SDL_isdigit SDL_isdigit_REAL
#define SDL_isgraph SDL_isgraph_REAL
#define SDL_isinf SDL_isinf_REAL
#define SDL_isinff SDL_isinff_REAL
#define SDL_islower SDL_islower_REAL
#define SDL_isnan SDL_isnan_REAL
#define SDL_isnanf SDL_isnanf_REAL
#define SDL_isprint SDL_isprint_REAL
#define SDL_ispunct SDL_ispunct_REAL
#define SDL_isspace SDL_isspace_REAL
#define SDL_isupper SDL_isupper_REAL
#define SDL_isxdigit SDL_isxdigit_REAL
#define SDL_itoa SDL_itoa_REAL
#define SDL_lltoa SDL_lltoa_REAL
#define SDL_log SDL_log_REAL
#define SDL_log10 SDL_log10_REAL
#define SDL_log10f SDL_log10f_REAL
#define SDL_logf SDL_logf_REAL
#define SDL_lround SDL_lround_REAL
#define SDL_lroundf SDL_lroundf_REAL
#define SDL_ltoa SDL_ltoa_REAL
#define SDL_malloc SDL_malloc_REAL
#define SDL_memcmp SDL_memcmp_REAL
#define SDL_memcpy SDL_memcpy_REAL
#define SDL_memmove SDL_memmove_REAL
#define SDL_memset SDL_memset_REAL
#define SDL_memset4 SDL_memset4_REAL
#define SDL_modf SDL_modf_REAL
#define SDL_modff SDL_modff_REAL
#define SDL_murmur3_32 SDL_murmur3_32_REAL
#define SDL_pow SDL_pow_REAL
#define SDL_powf SDL_powf_REAL
#define SDL_qsort SDL_qsort_REAL
#define SDL_qsort_r SDL_qsort_r_REAL
#define SDL_rand SDL_rand_REAL
#define SDL_rand_bits SDL_rand_bits_REAL
#define SDL_rand_bits_r SDL_rand_bits_r_REAL
#define SDL_rand_r SDL_rand_r_REAL
#define SDL_randf SDL_randf_REAL
#define SDL_randf_r SDL_randf_r_REAL
#define SDL_realloc SDL_realloc_REAL
#define SDL_round SDL_round_REAL
#define SDL_roundf SDL_roundf_REAL
#define SDL_scalbn SDL_scalbn_REAL
#define SDL_scalbnf SDL_scalbnf_REAL
#define SDL_setenv_unsafe SDL_setenv_unsafe_REAL
#define SDL_sin SDL_sin_REAL
#define SDL_sinf SDL_sinf_REAL
#define SDL_snprintf    SDL_snprintf_REAL
#define SDL_sqrt SDL_sqrt_REAL
#define SDL_sqrtf SDL_sqrtf_REAL
#define SDL_srand SDL_srand_REAL
#define SDL_sscanf  SDL_sscanf_REAL
#define SDL_strcasecmp SDL_strcasecmp_REAL
#define SDL_strcasestr SDL_strcasestr_REAL
#define SDL_strchr SDL_strchr_REAL
#define SDL_strcmp SDL_strcmp_REAL
#define SDL_strdup SDL_strdup_REAL
#define SDL_strlcat SDL_strlcat_REAL
#define SDL_strlcpy SDL_strlcpy_REAL
#define SDL_strlen SDL_strlen_REAL
#define SDL_strlwr SDL_strlwr_REAL
#define SDL_strncasecmp SDL_strncasecmp_REAL
#define SDL_strncmp SDL_strncmp_REAL
#define SDL_strndup SDL_strndup_REAL
#define SDL_strnlen SDL_strnlen_REAL
#define SDL_strnstr SDL_strnstr_REAL
#define SDL_strpbrk SDL_strpbrk_REAL
#define SDL_strrchr SDL_strrchr_REAL
#define SDL_strrev SDL_strrev_REAL
#define SDL_strstr SDL_strstr_REAL
#define SDL_strtod SDL_strtod_REAL
#define SDL_strtok_r SDL_strtok_r_REAL
#define SDL_strtol SDL_strtol_REAL
#define SDL_strtoll SDL_strtoll_REAL
#define SDL_strtoul SDL_strtoul_REAL
#define SDL_strtoull SDL_strtoull_REAL
#define SDL_strupr SDL_strupr_REAL
#define SDL_swprintf SDL_swprintf_REAL
#define SDL_tan SDL_tan_REAL
#define SDL_tanf SDL_tanf_REAL
#define SDL_tolower SDL_tolower_REAL
#define SDL_toupper SDL_toupper_REAL
#define SDL_trunc SDL_trunc_REAL
#define SDL_truncf SDL_truncf_REAL
#define SDL_uitoa SDL_uitoa_REAL
#define SDL_ulltoa SDL_ulltoa_REAL
#define SDL_ultoa SDL_ultoa_REAL
#define SDL_unsetenv_unsafe SDL_unsetenv_unsafe_REAL
#define SDL_utf8strlcpy SDL_utf8strlcpy_REAL
#define SDL_utf8strlen SDL_utf8strlen_REAL
#define SDL_utf8strnlen SDL_utf8strnlen_REAL
#define SDL_vasprintf SDL_vasprintf_REAL
#define SDL_vsnprintf SDL_vsnprintf_REAL
#define SDL_vsscanf SDL_vsscanf_REAL
#define SDL_vswprintf SDL_vswprintf_REAL
#define SDL_wcscasecmp SDL_wcscasecmp_REAL
#define SDL_wcscmp SDL_wcscmp_REAL
#define SDL_wcsdup SDL_wcsdup_REAL
#define SDL_wcslcat SDL_wcslcat_REAL
#define SDL_wcslcpy SDL_wcslcpy_REAL
#define SDL_wcslen SDL_wcslen_REAL
#define SDL_wcsncasecmp SDL_wcsncasecmp_REAL
#define SDL_wcsncmp SDL_wcsncmp_REAL
#define SDL_wcsnlen SDL_wcsnlen_REAL
#define SDL_wcsnstr SDL_wcsnstr_REAL
#define SDL_wcsstr SDL_wcsstr_REAL
#define SDL_wcstol SDL_wcstol_REAL
#define SDL_StepBackUTF8 SDL_StepBackUTF8_REAL
#define SDL_DelayPrecise SDL_DelayPrecise_REAL
#define SDL_CalculateGPUTextureFormatSize SDL_CalculateGPUTextureFormatSize_REAL
#define SDL_SetErrorV SDL_SetErrorV_REAL
#define SDL_GetDefaultLogOutputFunction SDL_GetDefaultLogOutputFunction_REAL
#define SDL_RenderDebugText SDL_RenderDebugText_REAL
#define SDL_GetSandbox SDL_GetSandbox_REAL
#define SDL_CancelGPUCommandBuffer SDL_CancelGPUCommandBuffer_REAL
#define SDL_SaveFile_IO SDL_SaveFile_IO_REAL
#define SDL_SaveFile SDL_SaveFile_REAL
#define SDL_GetCurrentDirectory SDL_GetCurrentDirectory_REAL
#define SDL_IsAudioDevicePhysical SDL_IsAudioDevicePhysical_REAL
#define SDL_IsAudioDevicePlayback SDL_IsAudioDevicePlayback_REAL
#define SDL_AsyncIOFromFile SDL_AsyncIOFromFile_REAL
#define SDL_GetAsyncIOSize SDL_GetAsyncIOSize_REAL
#define SDL_ReadAsyncIO SDL_ReadAsyncIO_REAL
#define SDL_WriteAsyncIO SDL_WriteAsyncIO_REAL
#define SDL_CloseAsyncIO SDL_CloseAsyncIO_REAL
#define SDL_CreateAsyncIOQueue SDL_CreateAsyncIOQueue_REAL
#define SDL_DestroyAsyncIOQueue SDL_DestroyAsyncIOQueue_REAL
#define SDL_GetAsyncIOResult SDL_GetAsyncIOResult_REAL
#define SDL_WaitAsyncIOResult SDL_WaitAsyncIOResult_REAL
#define SDL_SignalAsyncIOQueue SDL_SignalAsyncIOQueue_REAL
#define SDL_LoadFileAsync SDL_LoadFileAsync_REAL
#define SDL_ShowFileDialogWithProperties SDL_ShowFileDialogWithProperties_REAL
#define SDL_IsMainThread SDL_IsMainThread_REAL
#define SDL_RunOnMainThread SDL_RunOnMainThread_REAL
#define SDL_SetGPUAllowedFramesInFlight SDL_SetGPUAllowedFramesInFlight_REAL
#define SDL_RenderTextureAffine SDL_RenderTextureAffine_REAL
#define SDL_WaitForGPUSwapchain SDL_WaitForGPUSwapchain_REAL
#define SDL_WaitAndAcquireGPUSwapchainTexture SDL_WaitAndAcquireGPUSwapchainTexture_REAL
#define SDL_RenderDebugTextFormat SDL_RenderDebugTextFormat_REAL
#define SDL_CreateTray SDL_CreateTray_REAL
#define SDL_SetTrayIcon SDL_SetTrayIcon_REAL
#define SDL_SetTrayTooltip SDL_SetTrayTooltip_REAL
#define SDL_CreateTrayMenu SDL_CreateTrayMenu_REAL
#define SDL_CreateTraySubmenu SDL_CreateTraySubmenu_REAL
#define SDL_GetTrayMenu SDL_GetTrayMenu_REAL
#define SDL_GetTraySubmenu SDL_GetTraySubmenu_REAL
#define SDL_GetTrayEntries SDL_GetTrayEntries_REAL
#define SDL_RemoveTrayEntry SDL_RemoveTrayEntry_REAL
#define SDL_InsertTrayEntryAt SDL_InsertTrayEntryAt_REAL
#define SDL_SetTrayEntryLabel SDL_SetTrayEntryLabel_REAL
#define SDL_GetTrayEntryLabel SDL_GetTrayEntryLabel_REAL
#define SDL_SetTrayEntryChecked SDL_SetTrayEntryChecked_REAL
#define SDL_GetTrayEntryChecked SDL_GetTrayEntryChecked_REAL
#define SDL_SetTrayEntryEnabled SDL_SetTrayEntryEnabled_REAL
#define SDL_GetTrayEntryEnabled SDL_GetTrayEntryEnabled_REAL
#define SDL_SetTrayEntryCallback SDL_SetTrayEntryCallback_REAL
#define SDL_DestroyTray SDL_DestroyTray_REAL
#define SDL_GetTrayEntryParent SDL_GetTrayEntryParent_REAL
#define SDL_GetTrayMenuParentEntry SDL_GetTrayMenuParentEntry_REAL
#define SDL_GetTrayMenuParentTray SDL_GetTrayMenuParentTray_REAL
#define SDL_GetThreadState SDL_GetThreadState_REAL
#define SDL_AudioStreamDevicePaused SDL_AudioStreamDevicePaused_REAL
#define SDL_ClickTrayEntry SDL_ClickTrayEntry_REAL
#define SDL_UpdateTrays SDL_UpdateTrays_REAL
#define SDL_StretchSurface SDL_StretchSurface_REAL
